// -*- C++ -*-
/*!
 * @file  DataConversionRTC.cpp
 * @brief DataConversionRTC
 * @date $Date$
 *
 * $Id$
 */

#include "DataConversionRTC.h"

// Module specification
// <rtc-template block="module_spec">
static const char* dataconversionrtc_spec[] =
  {
    "implementation_id", "DataConversionRTC",
    "type_name",         "DataConversionRTC",
    "description",       "DataConversionRTC",
    "version",           "1.0.0",
    "vendor",            "TakahashiNaoki",
    "category",          "Three-WheeledMobileRobot",
    "activity_type",     "PERIODIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "1",
    "language",          "C++",
    "lang_type",         "compile",
    // Configuration variables
    "conf.default.PGainL", "5.0",
    "conf.default.DGainL", "2.0",
    "conf.default.PGainR", "5.0",
    "conf.default.DGainR", "2.0",
    "conf.default.leftWheelID", "1",
    "conf.default.rightWheelID", "0",
    "conf.default.radiusOfLeftWheel", "0.018",
    "conf.default.radiusOfRightWheel", "0.018",
    "conf.default.lengthOfAxle", "0.078",
    "conf.default.radiusOfBodyArea", "0.2",
    "conf.default.torqueSensitivity", "4.6",
    "conf.default.torqueConstant", "2.0",
    // Widget
    "conf.__widget__.PGainL", "text",
    "conf.__widget__.DGainL", "text",
    "conf.__widget__.PGainR", "text",
    "conf.__widget__.DGainR", "text",
    "conf.__widget__.leftWheelID", "text",
    "conf.__widget__.rightWheelID", "text",
    "conf.__widget__.radiusOfLeftWheel", "text",
    "conf.__widget__.radiusOfRightWheel", "text",
    "conf.__widget__.lengthOfAxle", "text",
    "conf.__widget__.radiusOfBodyArea", "text",
    "conf.__widget__.torqueSensitivity", "text",
    "conf.__widget__.torqueConstant", "text",
    // Constraints
    ""
  };
// </rtc-template>

/*!
 * @brief constructor
 * @param manager Maneger Object
 */
DataConversionRTC::DataConversionRTC(RTC::Manager* manager)
    // <rtc-template block="initializer">
  : RTC::DataFlowComponentBase(manager),
    m_model_angleIn("model_angle", m_model_angle),
    m_target_velocityIn("target_velocity", m_target_velocity),
    m_model_torqueOut("model_torque", m_model_torque),
    m_robot_commandOut("robot_command", m_robot_command)

    // </rtc-template>
{
}

/*!
 * @brief destructor
 */
DataConversionRTC::~DataConversionRTC()
{
}



RTC::ReturnCode_t DataConversionRTC::onInitialize()
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  addInPort("model_angle", m_model_angleIn);
  addInPort("target_velocity", m_target_velocityIn);
  
  // Set OutPort buffer
  addOutPort("model_torque", m_model_torqueOut);
  addOutPort("robot_command", m_robot_commandOut);
  
  // Set service provider to Ports
  
  // Set service consumers to Ports
  
  // Set CORBA Service Ports
  
  // </rtc-template>

  // <rtc-template block="bind_config">
  // Bind variables and configuration variable
  bindParameter("PGainL", m_PGainL, "5.0");
  bindParameter("DGainL", m_DGainL, "2.0");
  bindParameter("PGainR", m_PGainR, "5.0");
  bindParameter("DGainR", m_DGainR, "2.0");
  bindParameter("leftWheelID", m_leftWheelID, "1");
  bindParameter("rightWheelID", m_rightWheelID, "0");
  bindParameter("radiusOfLeftWheel", m_radiusOfLeftWheel, "0.018");
  bindParameter("radiusOfRightWheel", m_radiusOfRightWheel, "0.018");
  bindParameter("lengthOfAxle", m_lengthOfAxle, "0.078");
  bindParameter("radiusOfBodyArea", m_radiusOfBodyArea, "0.2");
  bindParameter("torqueSensitivity", m_torqueSensitivity, "4.6");
  bindParameter("torqueConstant", m_torqueConstant, "2.0");
  
  // </rtc-template>

  //Jn
  m_model_torque.data.length(Dof);	//Dof = 2
  m_robot_command.data.length(4);
  //I

  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t DataConversionRTC::onFinalize()
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t DataConversionRTC::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t DataConversionRTC::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/


RTC::ReturnCode_t DataConversionRTC::onActivated(RTC::UniqueId ec_id)
{
//MotorControl̃ASYėp@Jn

//---[add]---------------------------------
	RTC_INFO(("on Activated"));
	//std::cout << "on Activated" << std::endl;

	// Update INPORT 
	m_model_angleIn.update();
	m_target_velocityIn.update();

//--get Model Parameter from Configulation
	Body.leftID = m_leftWheelID; // jointID of left wheel
	Body.rightID = m_rightWheelID; // jointID of right wheel
	Body.wheelRadiusLeft = m_radiusOfLeftWheel; // Wheel Radius(Left)
	Body.wheelRadiusRight = m_radiusOfRightWheel; // Wheel Radius(Right)
	Body.bodyRadius = 0.5 * m_lengthOfAxle; // Body Radius (half of length of Wheel Axle)
	Body.length = m_radiusOfBodyArea; // length from center of Body to corner of Body

	RTC_DEBUG(("Body.leftID: %d   Body.rightID: %d", Body.leftID, Body.rightID));	
	RTC_DEBUG(("Body.wheelRadius[L]: %f Body.wheelRadius[R]: %f   Body.bodyRadius: %f",Body.wheelRadiusLeft, Body.wheelRadiusRight, Body.bodyRadius));	

  // set previous wheel angular velocity's data 
	for(int i=0; i < Dof; i++)  	qOld[i] = 0.0;  // initialize to current wheel's angle value

  // initialize SWITCH
	InitializeFlag = false;

	//-- initialize Time data
	sec = nsec = 0;	
	TimeStampOld = TimeStamp = 0.0;

//---[add]---------------------------------

//MotorControl̃ASYėp@I

	const char* output = "1kok";	//1kok:~
	memcpy(&(m_robot_command.data[0]),output,4);	//m_robot_command.data[0]="1"Cm_robot_command.data[2]="k"Cm_robot_command.data[2]="o"Cm_robot_command.data[3]="k"
	m_robot_commandOut.write();	//o̓|[gɏ

  return RTC::RTC_OK;
}


RTC::ReturnCode_t DataConversionRTC::onDeactivated(RTC::UniqueId ec_id)
{
	const char* output = "1kok";	//1kok:~
	memcpy(&(m_robot_command.data[0]),output,4);
	m_robot_commandOut.write();

  return RTC::RTC_OK;
}

RTC::ReturnCode_t DataConversionRTC::onExecute(RTC::UniqueId ec_id)
{
//MotorControl̃ASYėp@JniRgȂǒǉj

//---[add]---------------------------------
//  std::cout << " ==================================== [onExecute] ====================================" << std::endl;

/*********************************************************************************************************/
//  main action
/*********************************************************************************************************/

	if(m_model_angleIn.isNew()){	//{bgf̎ԗ։]px̓͂XVĂ邩mF
		//-- get Previous Wheel Angle Data 
		m_model_angleIn.read();	//{bgf̎ԗ։]px̓͂ǂݍ

		//-- set Previous Wheel Angle Data
		double q[Dof];
		q[Body.leftID] = m_model_angle.data[Body.leftID];	//{bgf̍ԗ։]px
		q[Body.rightID] = m_model_angle.data[Body.rightID];	//{bgf̉Eԗ։]px

		//-- get time stamp data 
		sec = m_model_angle.tm.sec;	//[sec]
		nsec = m_model_angle.tm.nsec;	//[nsec]

		TimeStamp = sec + 1.0e-9*nsec;	//^CX^v[sec]

		//-- initialize 
		if (!InitializeFlag) {	//s
			InitializeFlag = true;
		} else {
			//-- calc timestep value
			timeStep = TimeStamp - TimeStampOld;	//Oƍ̃^CX^v̍[s]  

			//-- calc Wheel Angular Velocity's Datas by previous angle data
			double dq[Dof];
			dq[Body.leftID] = (q[Body.leftID] - qOld[Body.leftID]) / timeStep;	//{bgf̍ԗ֊px̎Zo
			dq[Body.rightID] = (q[Body.rightID] - qOld[Body.rightID]) / timeStep;	//{bgf̉Eԗ֊px̎Zo

			//-- calc TORQUE  & set data to OUTPORT
			double q_ref[Dof], dq_ref[Dof];	//pxCpx

			//-- action if Target Velocity datas are given 
			if(m_target_velocityIn.isNew()){	//ڕWx̓͂XVĂ邩mF

				//*** read Target Velocity data
				m_target_velocityIn.read();	//ڕWx̓͂̓ǂݍ

				double TransVel,RotVel,Rho;
				TransVel = m_target_velocity.data.vx;	//ڕWix[m/s]
				RotVel = m_target_velocity.data.va;	//ڕW񑬓x[rad/s]

				if (RotVel == 0.0) { 	//ڕW񑬓x0.0[rad/s]Ȃ
					//-- calc [Target] Angular Velocity Dat
					dq_ref[Body.leftID] = TransVel / Body.wheelRadiusLeft * m_torqueConstant;	//ڕWpxiԗցj[rad/s]C* m_torqueConstant ǉ
					dq_ref[Body.rightID] = TransVel / Body.wheelRadiusRight * m_torqueConstant;	//ڕWpxiEԗցj[rad/s]C* m_torqueConstant ǉ
				} else {	//ڕW񑬓x0.0[rad/s]ȊOȂ
					Rho = TransVel / RotVel;        // curvature radius
					//-- calc [Target] Angular Velocity Dat
					dq_ref[Body.leftID] = (Rho - Body.bodyRadius) * RotVel / Body.wheelRadiusLeft * m_torqueConstant;	//ڕWpxiԗցj[rad/s]C* m_torqueConstant ǉ
					dq_ref[Body.rightID] = (Rho + Body.bodyRadius) * RotVel / Body.wheelRadiusRight * m_torqueConstant;	//ڕWpxiԗցj[rad/s]C* m_torqueConstant ǉ
				}

				//-- calc [Target] Angle Data
				q_ref[Body.leftID] = q[Body.leftID] + 0.5 * (dq_ref[Body.leftID] + dq[Body.leftID]) * timeStep;	//ڕWԗ։]pxiԗցj[rad]
				q_ref[Body.rightID] = q[Body.rightID] + 0.5 * (dq_ref[Body.rightID] + dq[Body.rightID]) * timeStep;	//ڕWԗ։]pxiEԗցj[rad]

				//-- calc TORQUE value    
				m_model_torque.data[Body.leftID] = ( (q_ref[Body.leftID] - q[Body.leftID]) * m_PGainL + (dq_ref[Body.leftID] - dq[Body.leftID]) * m_DGainL ) * m_torqueSensitivity;	//ԗւɗ^gNZoC* m_torqueSensitivity ǉ
				m_model_torque.data[Body.rightID] = ( (q_ref[Body.rightID] - q[Body.rightID]) * m_PGainR + (dq_ref[Body.rightID] - dq[Body.rightID]) * m_DGainR ) * m_torqueSensitivity;	//Eԗւɗ^gNZoC* m_torqueSensitivity ǉ

				//OɃf[^ۑ
				RTC_DEBUG((" Pgain(L): %f Pgain(R): %f Dgain(L): %f Dgain(R): %f", m_PGainL, m_PGainR, m_DGainL, m_DGainR));;
				RTC_DEBUG((" TransVel: %f RotVel: %f", TransVel, RotVel));
				RTC_DEBUG((" q_ref(L): %f q(L): %f [q_ref-q](L): %f dq_ref(L): %f dq(L): %f [dq_ref-dq](L): %f", q_ref[Body.leftID], q[Body.leftID], q_ref[Body.leftID] - q[Body.leftID], dq_ref[Body.leftID], dq[Body.leftID], dq_ref[Body.leftID] - dq[Body.leftID]));	
				RTC_DEBUG((" q_ref(R): %f q(R): %f [q_ref-q](R): %f dq_ref(R): %f dq(R): %f [dq_ref-dq](R): %f", q_ref[Body.rightID], q[Body.rightID], q_ref[Body.rightID] - q[Body.rightID], dq_ref[Body.rightID], dq[Body.rightID], dq_ref[Body.rightID] - dq[Body.rightID]));
				RTC_DEBUG((" Ptorque(L): %f Dtorque(L): %f", (q_ref[Body.leftID] - q[Body.leftID]) * m_PGainL, (dq_ref[Body.leftID] - dq[Body.leftID]) * m_DGainL));
				RTC_DEBUG((" Ptorque(R): %f Dtorque(R): %f", (q_ref[Body.rightID] - q[Body.rightID]) * m_PGainR, (dq_ref[Body.rightID] - dq[Body.rightID]) * m_DGainR));

//MotorControl̃ASYėp@I

/*********************************************************************************************************/
//  @̓ݒ@Jn
/*********************************************************************************************************/

				const char* output;
				double RotVel_deg = RotVel * 180 / 3.141592;	//1bŉ]px[deg]Cv肪
	
				//ix[
				if(TransVel == 0.00 && RotVel_deg == 0.00){output = "1kok";}	//1kok:~
				//ixvX
				else if(0.00 <= TransVel && TransVel < 0.225){
					//1bŉ]px-90[deg]90[deg]͈̔͂5
					if(-90 <= RotVel_deg && RotVel_deg < -63){output = "5dok";}	//Œᑬxŋ}ȉE܁iE񂩂̑Oij
					else if(-63 <= RotVel_deg && RotVel_deg < -36){output = "4dok";}	//ᑬŋ}ȉE
					else if(-36 <= RotVel_deg && RotVel_deg < 36){output = "5aok";}	//ŒᑬxőSg
					else if(36 <= RotVel_deg && RotVel_deg < 63){output = "4fok";}	//ᑬŋ}ȍ
					else if(63 <= RotVel_deg && RotVel_deg <= 90){output = "5fok";}	//Œᑬxŋ}ȍ܁i񂩂̑Oij
				}
				else if(0.225 <= TransVel){
					//1bŉ]px-90[deg]90[deg]͈̔͂5
					if(-90 <= RotVel_deg && RotVel_deg < -63){output = "4dok";}	//ᑬŋ}ȉE
					else if(-63 <= RotVel_deg && RotVel_deg < -36){output = "4cok";}	//ᑬŉE
					else if(-36 <= RotVel_deg && RotVel_deg < 36){output = "4aok";}	//ᑬőOi
					else if(36 <= RotVel_deg && RotVel_deg < 63){output = "4eok";}	//ᑬō
					else if(63 <= RotVel_deg && RotVel_deg <= 90){output = "4fok";}	//ᑬŋ}ȍ
				}
				//ix}CiX
				else if(0.00 > TransVel && TransVel > -0.225){
					//1bŉ]px-90[deg]90[deg]͈̔͂5
					if(-90 <= RotVel_deg && RotVel_deg < -63){output = "5jok";}	//Œᑬxŋ}ȍށi񂩂̌ށj
					else if(-63 <= RotVel_deg && RotVel_deg < -36){output = "4jok";}	//ᑬŋ}ȍ
					else if(-36 <= RotVel_deg && RotVel_deg < 36){output = "5bok";}	//ᑬŌ
					else if(36 <= RotVel_deg && RotVel_deg < 63){output = "4hok";}	//ᑬŋ}ȉE
					else if(63 <= RotVel_deg && RotVel_deg <= 90){output = "5hok";}	//Œᑬxŋ}ȉEށiE񂩂̌ށj
				}
				else if(-0.225 >= TransVel){
					//1bŉ]px-90[deg]90[deg]͈̔͂5
					if(-90 <= RotVel_deg && RotVel_deg < -63){output = "4jok";}	//ᑬŋ}ȍ
					else if(-63 <= RotVel_deg && RotVel_deg < -36){output = "4iok";}	//ᑬō
					else if(-36 <= RotVel_deg && RotVel_deg < 36){output = "4bok";}	//ᑬŌ
					else if(36 <= RotVel_deg && RotVel_deg < 63){output = "4gok";}	//ᑬŉE
					else if(63 <= RotVel_deg && RotVel_deg <= 90){output = "4hok";}	//ᑬŋ}ȉE
				}

				memcpy(&(m_robot_command.data[0]),output,4);
				m_robot_commandOut.write();
				std::cout << "m_robot_command.data[0][1][2][3] = " << m_robot_command.data[0] << m_robot_command.data[1] << m_robot_command.data[2] << m_robot_command.data[3] << std::endl;	//^[~iɎwߒl\

/*********************************************************************************************************/
//  @̓ݒ@I
/*********************************************************************************************************/

//MotorControl̃ASYėp@Jn

			} else {	//ڕWixXVĂȂΗ^gN0.0
				m_model_torque.data[Body.leftID] = 0.0;
				m_model_torque.data[Body.rightID] = 0.0;
			}

	RTC_DEBUG(("[MotorControl Torque] Time = %f[s]  Torque(L,R) = (%f, %f)", TimeStamp, m_model_torque.data[Body.leftID], m_model_torque.data[Body.rightID]));

			// set TIME
			m_model_torque.tm.sec = sec;
			m_model_torque.tm.nsec = nsec;
			//*** set Torque data to OutPort
			m_model_torqueOut.write();	//o̓|[gɏ

		} // enod of if (initialize)

	//*** set Current data -> Previous data
		qOld[Body.leftID] = q[Body.leftID];
		qOld[Body.rightID] = q[Body.rightID];
		TimeStampOld = TimeStamp;

	}

//---[add]---------------------------------

//MotorControl̃ASYėp@I

  return RTC::RTC_OK;
}
/*
RTC::ReturnCode_t DataConversionRTC::onAborting(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t DataConversionRTC::onError(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t DataConversionRTC::onReset(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t DataConversionRTC::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t DataConversionRTC::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/



extern "C"
{
 
  void DataConversionRTCInit(RTC::Manager* manager)
  {
    coil::Properties profile(dataconversionrtc_spec);
    manager->registerFactory(profile,
                             RTC::Create<DataConversionRTC>,
                             RTC::Delete<DataConversionRTC>);
  }
  
};


